# ctle Copyright (c) 2023 Ulrik Lindahl
# Licensed under the MIT license https://github.com/Cooolrik/ctle/blob/main/LICENSE

from enum import Enum
from .formatted_output import formatted_output

class enum_value:
    def __init__( self, name:str, value:str, desc:str ) -> None:
        self.name = name
        self.value = value
        self.desc = ' '.join(desc.split()) # removes all consecutive whitespace and replaces with just a single space

class enum_type:
    def __init__( self, name:str, types:str, desc:str, values:list[enum_value] ) -> None:
        self.name = name
        self.types = types
        self.desc = desc
        self.values = values
    
# name_padding - minimum number of chars in name, (will be right-padded with spaces if less)
# value_padding - minimum number of chars in value, (will be right-padded with spaces if less)
# pad_to_max - automatically increase paddings based on names and values in the enum
def generate_enum_declaration( enm:enum_type, out:formatted_output, name_padding:int = 0, value_padding:int = 0, pad_to_max:bool = True  ) -> None:
    if pad_to_max:
        for val in enm.values:
            name_padding = max( name_padding, len(val.name) )
            value_padding = max( value_padding, len(val.value) )

    if enm.desc != None and len(enm.desc) > 0:
        out.comment_ln(enm.desc)

    s = 'enum class ' + enm.name
    if enm.types != None and len(enm.types) > 0:
        s += ' : ' + enm.types
    out.ln(s)

    with out.blk( add_semicolon=True ):
        for val in enm.values:
            # get name, pad with spaces
            s = val.name
            if len(val.name) < name_padding:
                s += ' ' * (name_padding - len(val.name)) 
            # add value, pad with spaces
            if val.value != None:
                s += ' = '
                if len(val.value) < value_padding:
                    s += ' ' * (value_padding - len(val.value)) 
                s += val.value
            # end of value, add comment
            s += ','
            if val.desc != None and len(val.desc) > 0:
                s += ' // ' + val.desc
            out.ln(s)
    
def generate_enum_function_declarations( enm:enum_type, out:formatted_output ) -> None:
    out.comment_ln('Converts a string into an enum: ' + enm.name + ' value.' )
    out.comment_ln('Returns status::ok on successful conversion.' )
    out.ln(f'ctle::status convert_string_to_enum( {enm.name} &dest , const std::string &src );')
    out.ln(f'ctle::status convert_string_to_enum( {enm.name} &dest , const char *src );')
    out.ln()

    out.comment_ln('Write the enum: ' + enm.name + ' value into a string. Checks that the enum value is valid, and returns status::ok on success.' )
    out.comment_ln('The version which retuns a string, returns an empty string if the enum is not a valid value.' )
    out.ln(f'ctle::status convert_enum_to_string( std::string &dest , {enm.name} src );')
    out.ln(f'std::string convert_enum_to_string( {enm.name} src );')
    out.ln()

    out.comment_ln('Get the description of the enum: ' + enm.name + ' value into as a string. Returns status::ok if a description is found.' )
    out.comment_ln('The version which retuns a string, returns an empty string if a description is not found.' )
    out.ln(f'ctle::status get_enum_value_description( std::string &dest , {enm.name} src );')
    out.ln(f'std::string get_enum_value_description( {enm.name} src );')
    out.ln()

    out.comment_ln('Checks if the value is a valid enum: ' + enm.name + ' and returns true if the value checks out ok.' )
    out.ln(f'bool is_valid_enum( {enm.name} value );')
    out.ln()    

# if use_lowercase_strings is True, all strings will be generated as lowercase. 
# if use_lowercase_strings is False, the strings will be generated identical as the enum values
def generate_enum_function_implementations( enm:enum_type, out:formatted_output, use_lowercase_strings:bool = False ) -> None:
    
    # generate the look-up tables 
    out.ln(f'static const std::unordered_map<{enm.name}, const char *> enum_{enm.name}_to_string_mapping = ')
    with out.blk( add_semicolon=True ):
        for val in enm.values:
            if use_lowercase_strings:
                out.blk_ln(f'{enm.name}::{val.name} , "{val.name.lower()}"', add_comma=True )
            else:
                out.blk_ln(f'{enm.name}::{val.name} , "{val.name}"', add_comma=True )
    out.ln()

    out.ln(f'static const std::unordered_map<{enm.name}, const char *> enum_{enm.name}_to_description_mapping = ')
    with out.blk( add_semicolon=True ):
        for val in enm.values:
            if val.desc != None:
                out.blk_ln(f'{enm.name}::{val.name} , "{val.desc}"', add_comma=True )
    out.ln()    

    out.ln(f'static const std::unordered_map<std::string, {enm.name}> enum_string_to_{enm.name}_mapping = ')
    with out.blk( add_semicolon=True ):
        for val in enm.values:
            if use_lowercase_strings:
                out.blk_ln(f'"{val.name.lower()}" , {enm.name}::{val.name}', add_comma=True )
            else:
                out.blk_ln(f'"{val.name}" , {enm.name}::{val.name}', add_comma=True )
    out.ln()   

    out.ln(f'ctle::status convert_string_to_enum( {enm.name} &dest , const std::string &src )')
    with out.blk():
        out.comment_ln('look up value, check if found')
        out.ln(f'auto it = enum_string_to_{enm.name}_mapping.find(src);')
        out.ln(f'if( it == enum_string_to_{enm.name}_mapping.end() )')
        with out.blk():
            out.ln('return ctle::status::not_found;')
        out.ln()
        out.comment_ln('all ok, return value in dest')
        out.ln('dest = it->second;')
        out.ln('return ctle::status::ok;')
    out.ln()

    out.ln(f'ctle::status convert_string_to_enum( {enm.name} &dest , const char *src )')
    with out.blk():
        out.comment_ln('check validity of input param')
        out.ln('if( !src )')
        with out.blk():
            out.ln('return ctle::status::invalid_param;')
        out.ln()
        out.comment_ln('look up value using std::string version of the function')
        out.ln('return convert_string_to_enum( dest, std::string(src) );')
    out.ln()

    out.ln(f'ctle::status convert_enum_to_string( std::string &dest , {enm.name} src )')
    with out.blk():
        out.comment_ln('look up enum value, check if found')
        out.ln(f'auto it = enum_{enm.name}_to_string_mapping.find(src);')
        out.ln(f'if( it == enum_{enm.name}_to_string_mapping.end() )')
        with out.blk():
            out.ln('return ctle::status::not_found;')
        out.ln()
        out.comment_ln('all ok, return name in dest')
        out.ln('dest = std::string(it->second);')
        out.ln('return ctle::status::ok;')    
    out.ln()

    out.ln(f'std::string convert_enum_to_string( {enm.name} src )')
    with out.blk():
        out.comment_ln('look up enum value, check if found')
        out.ln(f'auto it = enum_{enm.name}_to_string_mapping.find(src);')
        out.ln(f'if( it == enum_{enm.name}_to_string_mapping.end() )')
        with out.blk():
            out.ln('return ""; // not found')
        out.ln()
        out.comment_ln('all ok, return name string')
        out.ln('return std::string(it->second);')    
        out.ln()

    out.ln(f'ctle::status get_enum_value_description( std::string &dest , {enm.name} src )')
    with out.blk():
        out.comment_ln('look up enum value, check if found')
        out.ln(f'auto it = enum_{enm.name}_to_description_mapping.find(src);')
        out.ln(f'if( it == enum_{enm.name}_to_description_mapping.end() )')
        with out.blk():
            out.ln('return ctle::status::not_found;')
        out.ln()
        out.comment_ln('all ok, return description in dest')
        out.ln('dest = std::string(it->second);')
        out.ln('return ctle::status::ok;')    
    out.ln()

    out.ln(f'std::string get_enum_value_description( {enm.name} src )')
    with out.blk():
        out.comment_ln('look up enum value, check if found')
        out.ln(f'auto it = enum_{enm.name}_to_description_mapping.find(src);')
        out.ln(f'if( it == enum_{enm.name}_to_description_mapping.end() )')
        with out.blk():
            out.ln('return ""; // no description found')
        out.ln()
        out.comment_ln('all ok, return description string')
        out.ln('return std::string(it->second);')    
    out.ln()

    out.ln(f'bool is_valid_enum( {enm.name} value )')
    with out.blk():
        out.comment_ln('look up enum value, check if found')
        out.ln(f'auto it = enum_{enm.name}_to_string_mapping.find(value);')
        out.ln(f'if( it == enum_{enm.name}_to_string_mapping.end() )')
        with out.blk():
            out.ln('return false; // not found, not valid')
        out.ln()
        out.comment_ln('enum found')
        out.ln('return true;')    
    out.ln()




# ctle Copyright (c) 2023 Ulrik Lindahl
# Licensed under the MIT license https://github.com/Cooolrik/ctle/blob/main/LICENSE

import copy
import os
from stat import S_IRUSR, S_IRGRP, S_IROTH, S_IWUSR
import importlib
import textwrap
from contextlib import contextmanager

class formatted_output:
	'''formatted_output defined a class which can be used to more easily generate formatted output of C/C++ files. The class supports different kinds of indentation, and wraps comments.'''

	def __init__(self, indentation:int = 0, tab_str:str = '\t', indent_braces:bool = False, comment_wrap:int = 120 ) -> None:
		self.lines = []
		self.indent_braces = indent_braces
		self.indentation = indentation
		self.tab_str = tab_str
		self.comment_wrap = comment_wrap

	@contextmanager
	def blk(self, add_semicolon:bool = False, no_indent = False):
		'''Add an indented block of code'''
		try:
			st = self.tab_str * self.indentation
			if self.indent_braces:
				st += self.tab_str
			st += '{'
			self.lines.append(st)
			if not no_indent:
				self.indentation += 1	
			yield self

		finally:
			if not no_indent:
				self.indentation -= 1
			st = self.tab_str * self.indentation
			if self.indent_braces:
				st += self.tab_str 
			st += '}'
			if add_semicolon:
				st += ';'
			self.lines.append(st)

	@contextmanager
	def tab(self, prefix_ln:str = None):
		'''Add an indented block of code, only indent, no braces'''
		try:
			if prefix_ln:
				self.ln(prefix_ln)
			self.indentation += 1	
			yield self

		finally:
			self.indentation -= 1

	def ln(self, lin:str = None, no_indent = False, append = False, intent_adjust = 0 ):
		'''Add a single line, at current indentation, (or no indentation if no_indent is True).'''
		if append:
			if lin:
				self.lines[-1] += lin
		else:				
			if lin != None:
				if no_indent:
					self.lines.append( lin )
				else:
					self.lines.append( self.tab_str * (self.indentation + intent_adjust) + lin )
			else:
				self.lines.append('')
   
	# single line block
	def blk_ln(self, lin:str, add_comma:bool = False ):
		'''Add a single-line block, at the current indentation. Optionally add a comma at the end.'''
		st = ''
		if self.indent_braces:
			st += self.tab_str
		st += '{ ' + lin + ' }'
		if add_comma:
			st += ','
		self.ln( st )

	def comment_ln(self, comm:str):
		'''Add a commend, which is wrapped if longer than the comment_wrap length'''
		lines = textwrap.wrap( comm , width = self.comment_wrap )
		for lin in lines:
			self.ln('// ' + textwrap.dedent(lin) )

	def generate_license_header(self, project_name:str='ctle', copyright_holder:str = '2022 Ulrik Lindahl', license_type:str = 'MIT', license_link:str = 'https://github.com/Cooolrik/ctle/blob/main/LICENSE' ):
		'''Generates the license header and injects into the generated c++ file'''
		self.comment_ln(f'{project_name} Copyright (c) {copyright_holder}')
		self.comment_ln(f'Licensed under the {license_type} license {license_link}')

	def inline_file(self, path:str):
		'''Injects the contents of another file, indented at the current indentation level'''
		inlined_file = open(path, 'r')
		lines = inlined_file.readlines()
		for str in lines:
			self.ln( str.rstrip() )
		inlined_file.close()

	def generate_push_and_disable_warnings(self, vswarns:list[str] , gccwarns:list[str] , comment:str = 'Disable warnings that are caused by code which we cannot control' ):
		'''Insert a block of warning disable pragmas for VC and GCC. Pushes all warnings before disabling.'''
		self.comment_ln(comment)
		self.ln('#ifdef _MSC_VER')
		self.ln('#pragma warning( push )')
		for w in vswarns:
			self.ln(f'#pragma warning( disable : {w} )')
		self.ln('#elif defined(__GNUC__)')
		self.ln('#pragma GCC diagnostic push')
		for w in gccwarns:
			self.ln(f'#pragma GCC diagnostic ignored "{w}"')	
		self.ln('#endif')

	def generate_pop_warnings(self, comment:str = 'Re-enable warnings again' ):
		'''Insert a block which pops the disabled warnings for VC or GCC'''
		self.comment_ln(comment)
		self.ln('#ifdef _MSC_VER')
		self.ln('#pragma warning( pop )')
		self.ln('#elif defined(__GNUC__)')
		self.ln('#pragma GCC diagnostic pop')
		self.ln('#endif')

	def write_lines_to_file( self, path:str ):
		'''Write the formatted text to a file, but only if the file is not already exactly identical to the formatted text'''

		# make into one long string
		new_text = ''
		for line in self.lines:
			new_text += line + '\n'

		# if the file already exists, check if we have an update
		if os.path.exists(path):
			# try reading in the current file if one exists
			with open(path,'r') as f:
				existing_text = f.read()
				f.close()

			#if no difference was found, return
			if new_text == existing_text:
				print( '\tSkipping: ' + path + ', (it is identical)...')
				return

			# if a difference was found, remove the old file
			os.chmod(path, S_IWUSR)
			os.remove( path ) 

		# we should write to the new file
		print( '\tWriting: ' + path + '...')
		with open(path,'w') as f:
			f.write(new_text)
			f.close()

		# change mode of file to readonly
		os.chmod(path, S_IRUSR|S_IRGRP|S_IROTH)


